#include "SDL.h"
#include "SDL_opengl.h"
#include <math.h>
#include <string.h>


/* Player ball + logo */
#define PARTICLES 23

#define DETAIL 64
#define W 10
#define H 10
#define R 0.6f
#define D (R * 2)
#define R_SQ (R * R)
#define D_SQ (D * D)
#define BOUNCINESS 0.5f /* used only for edges */
#define DAMPING 0.992f

float gravity[3] = { 0, -0.02f, 0 };
Uint32 t = 0, startTime;
float particlePos[PARTICLES * 3], oldParticlePos[PARTICLES * 3];
float angle;
GLuint sphereList;
char falling[PARTICLES];

void restart()
{
    /* 22 balls from logo:
       o  o   o  o  o
       o o o o  o  o
       o o o o  o  o
       o  o   o  o  o
    */
    int a, b, p = 0, i[4] = { 0x2449, 0x2a92, 0x2a92, 0x2449 };
    particlePos[p++] = 0;
    particlePos[p++] = 1;
    particlePos[p++] = -5;
    for (b = 0; b < 4; ++b)
    {
        for (a = 0; a < 14; ++a)
        {
            if ((i[b] >> a) & 1)
            {
                particlePos[p++] = (float)(a - 6);
                particlePos[p++] = (float)b + 2;
                particlePos[p++] = (float)b * 4 / 3 + 3;
            }
        }
    }
    memcpy(oldParticlePos, particlePos, PARTICLES * 3 * sizeof(float));
    memset(falling, 0, PARTICLES);
    angle = -3.142f / 2;
}

void setup()
{
    GLUquadric *qobj;
    
    SDL_Init(32);
    SDL_SetVideoMode(640, 480, 32, 2);
    startTime = SDL_GetTicks();

    glClearColor(0.1f, 0.2f, 0.3f, 1);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);
    
    qobj = gluNewQuadric();
    gluQuadricNormals(qobj, GLU_SMOOTH);
    sphereList = glGenLists(2);
    glNewList(sphereList, GL_COMPILE);
    gluSphere(qobj, R, DETAIL, DETAIL / 2);
    glEndList();
    glNewList(sphereList + 1, GL_COMPILE);
    gluSphere(qobj, R, DETAIL / 2, DETAIL / 4);
    glEndList();
    gluDeleteQuadric(qobj);

    restart();
}


void update()
{
    int a;

    for (a = 0; a < PARTICLES * 3; ++a) {
        /* integrate particle pos */
        float t = particlePos[a];
        /* simplified verlet integration (no external forces except static gravity) */
        particlePos[a] += DAMPING * (particlePos[a] - oldParticlePos[a]) + gravity[a % 3];
        oldParticlePos[a] = t;
    }
    
    /* collisions between particles */
    for (a = 0; a < PARTICLES - 1; ++a)
    {
        int b;
        float *pa = &particlePos[a * 3];
        for (b = a + 1; b < PARTICLES; ++b)
        {
            float *pb = &particlePos[b * 3];
            float distSq = (pb[0] - pa[0]) * (pb[0] - pa[0]) +
                           (pb[1] - pa[1]) * (pb[1] - pa[1]) +
                           (pb[2] - pa[2]) * (pb[2] - pa[2]);
            if (distSq < D_SQ)
            {
                /* solve constraint between particles p1 and p2 */
                float dx = pb[0] - pa[0];
                float dy = pb[1] - pa[1];
                float dz = pb[2] - pa[2];
                /* sqrt approximation from Jakobsen character physics paper */
                float diff = D_SQ / (distSq + D_SQ) - 0.5f;
                dx *= diff; dy *= diff; dz *= diff;
                pa[0] -= dx; pa[1] -= dy; pa[2] -= dz;
                pb[0] += dx; pb[1] += dy; pb[2] += dz;
            }
        }
    }
    
    for (a = 0; a < PARTICLES; ++a)
    {
        float *p = &particlePos[a * 3];
        float *po = &oldParticlePos[a * 3];
        float bagDistSq;
        /* keep particle inside screen rect and bounce off from walls
           (particle speed is adjusted by bounciness by tweaking the
           old particle position to match the adjusted speed)
        */
        if (p[0] < -W) {
            float s = BOUNCINESS * (p[0] - po[0]);
            p[0] = -W;
            po[0] = p[0] + s;
        }
        if (p[2] < -H) {
            float s = BOUNCINESS * (p[2] - po[2]);
            p[2] = -H;
            po[2] = p[2] + s;
        }
        if (p[0] > W) {
            float s = BOUNCINESS * (p[0] - po[0]);
            p[0] = W;
            po[0] = p[0] + s;
        }
        if (p[2] > H) {
            float s = BOUNCINESS * (p[2] - po[2]);
            p[2] = H;
            po[2] = p[2] + s;
        }
        bagDistSq = p[0] * p[0] + p[2] * p[2];
        if (bagDistSq < D_SQ && p[1] < 0.01f)
            falling[a] = 1;
        if (p[1] < 0 && !falling[a]) {
            float s = BOUNCINESS * (p[1] - po[1]);
            p[1] = 0;
            po[1] = s;
        }
        if (falling[a])
        {
            p[0] *= 0.95f;
            p[2] *= 0.95f;
        }
    }
    if (falling[0] && particlePos[1] < -6)
        restart();
}



void box(float x, float y, float z)
{
    x /= 2;
    y /= 2;
    z /= 2;
    
    glBegin(GL_QUADS);
    /* Front face */
    glNormal3f(0, 0, 1);
    glVertex3f(-x, -y, z);
    glVertex3f(x, -y, z);
    glVertex3f(x, y, z);
    glVertex3f(-x, y, z);
    /* Back face */
    glNormal3f(0, 0, -1);
    glVertex3f(-x, -y, -z);
    glVertex3f(-x, y, -z);
    glVertex3f(x, y, -z);
    glVertex3f(x, -y, -z);
    /* Top face */
    glNormal3f(0, 1, 0);
    glVertex3f(-x, y, -z);
    glVertex3f(-x, y, z);
    glVertex3f(x, y, z);
    glVertex3f(x, y, -z);
    /* Bottom face */
    /*
    glNormal3f(0, -1, 0);
    glVertex3f(-x, -y, -z);
    glVertex3f(x, -y, -z);
    glVertex3f(x, -y, z);
    glVertex3f(-x, -y, z);
    */
    /* Left face */
    glNormal3f(-1, 0, 0);
    glVertex3f(-x, -y, -z);
    glVertex3f(-x, -y, z);
    glVertex3f(-x, y, z);
    glVertex3f(-x, y, -z);
    /* Right face */
    glNormal3f(1, 0, 0);
    glVertex3f(x, -y, -z);
    glVertex3f(x, y, -z);
    glVertex3f(x, y, z);
    glVertex3f(x, -y, z);
    glEnd();
}


void draw()
{
    int a;
    float light1Pos[4] = { 0, 5, 0, 0.3f };
    float light2Pos[4] = { 0, -1, 0, 0 };
    float light1Color[4] = { 0.6f, 0.9f, 1, 1 };
    float light2Color[4] = { 0.3f, 0.2f, 0.1f, 1 };
    float ambientColor[4] = { 0.25f, 0.25f, 0.25f, 1 };
    float specularReflection[4] = { 0.5f, 0.5f, 0.5f, 1 };
    float specularColor[4] = { 1, 1, 1, 1 };

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45, 1.3, 0.1, 50);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glEnable(GL_LIGHTING);

    glEnable(GL_LIGHT0);
    glEnable(GL_LIGHT1);
    glLightfv(GL_LIGHT0, GL_POSITION, light1Pos);
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambientColor);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light1Color);
    glLightfv(GL_LIGHT0, GL_SPECULAR, specularColor);
    glLightfv(GL_LIGHT1, GL_POSITION, light2Pos);
    //glLightfv(GL_LIGHT1, GL_AMBIENT, ambientColor);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, light2Color);
    glMaterialfv(GL_FRONT, GL_SPECULAR, specularReflection);
    glMateriali(GL_FRONT, GL_SHININESS, 64);

    gluLookAt(particlePos[0] + cos(angle) * 10,
              particlePos[1] + 5,
              particlePos[2] + sin(angle) * 10,
              particlePos[0], particlePos[1], particlePos[2],
              0, 1, 0);

    glColorMaterial(GL_FRONT, GL_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);

    glColor3d(0.4, 0.4, 0.6);
    glPushMatrix(); glTranslatef(-11, 0, 0); box(1, 2, 21); glPopMatrix();
    glPushMatrix(); glTranslatef(11, 0, 0); box(1, 2, 21); glPopMatrix();
    glColor3d(0.3, 0.3, 0.4);
    glPushMatrix(); glTranslatef(0, 0, -11); box(21, 2, 1); glPopMatrix();
    glPushMatrix(); glTranslatef(0, 0, 11); box(21, 2, 1); glPopMatrix();
    glColor3d(0.6, 0.3, 0.2);
    glPushMatrix(); glTranslatef(0, -1.6f, 0); box(21, 2, 21); glPopMatrix();
    
    glColor3f(1, 1, 1);
    for (a = 0; a < PARTICLES; ++a)
    {
        float *p = &particlePos[a * 3];
        if (!falling[a] || p[1] > -R)
        {
            glPushMatrix();
            glTranslatef(p[0], p[1], p[2]);
            glCallList(sphereList);
            glPopMatrix();
        }
        if (a == 0)
            glColor3d(0.4, 0.8, 0.4);
    }
    glDisable(GL_COLOR_MATERIAL);
    
    glDisable(GL_LIGHTING);
    glTranslated(0, -R + 0.01, 0);
    glScalef(1, 0, 1);
    glEnable(GL_BLEND);
    glBlendFunc(GL_DST_COLOR, GL_ZERO);
    glColor3d(0.2, 0.2, 0.2);
    glPushMatrix();
    glScaled(2, 0, 2);
    glCallList(sphereList + 1);
    glPopMatrix();
    glPushMatrix();
    glColor3d(0.7, 0.7, 0.7);
    for (a = 0; a < PARTICLES; ++a)
    {
        float *p = &particlePos[a * 3];
        if (falling[a])
            continue;
        glPushMatrix();
        glTranslatef(p[0], 0, p[2]);
        glScaled(0.9, 0, 0.9);
        glCallList(sphereList + 1);
        glPopMatrix();
    }
    glPopMatrix();
    glDisable(GL_BLEND);
    glEnable(GL_LIGHTING);
}


int main(int argc, char **argv)
{
    int alive = 1;
    int mouseDown = 0;
    argc = argc;
    argv = argv;

    setup();

    while (alive)
    {
        SDL_Event e;
        Uint32 currentTime = SDL_GetTicks() - startTime;
        while (SDL_PollEvent(&e))
        {
            int et = e.type;
            if (et == SDL_QUIT)
                alive = 0;
            if (et == SDL_KEYDOWN)
            {
                int key = e.key.keysym.sym;
                if (key == SDLK_ESCAPE)
                    alive = 0;
                if (key == SDLK_r)
                    restart();
                if (key == SDLK_SPACE || key == SDLK_u)
                {
                    particlePos[0] += (float)(cos(angle + 3.142) / 3);
                    particlePos[1] += ((key == SDLK_u) & 1) / 5.f;
                    particlePos[2] += (float)(sin(angle + 3.142) / 3);
                }
            }
            if (et == SDL_MOUSEBUTTONDOWN)
                mouseDown = 1;
            if (et == SDL_MOUSEBUTTONUP)
                mouseDown = 0;
            if (et == SDL_MOUSEMOTION && mouseDown)
                angle += e.motion.xrel / 100.f;
        }
        while (t < currentTime)
        {
            update();
            t += 1000 / 60;
        }
        draw();
        SDL_GL_SwapBuffers();
    }

    return 0;
}
